////////////////////////////////////////////////////////////////////////////////
// MinGW-w64 direct X 7 - ddraw test
////////////////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <tchar.h>
#include <D2D1.h>
#include <dwrite.h>
#include <wincodec.h>
#include <mmsystem.h>
#include <dsound.h>

#include "winwav.h"

#ifdef DEBUG
    #include <stdio.h>
    #include <stdlib.h>
    #include <string.h>
#endif

using namespace D2D1;

/*
#ifndef SafeRelease
    #define SafeRelease( _X_ ) { if ( _X_ != NULL ) { delete _X_; _X_ = NULL; } }
#endif /// SaveRelease
*/

template <class T> void SafeRelease(T **ppT)
{
    if (*ppT)
    {
        (*ppT)->Release();
        *ppT = NULL;
    }
}

////////////////////////////////////////////////////////////////////////////////

ID2D1Factory*                   pD2DFactory = NULL;
#ifdef USING_DC
ID2D1DCRenderTarget*            pD2DTarget   = NULL;
#else
ID2D1HwndRenderTarget*          pD2DTarget   = NULL;
#endif

ID2D1SolidColorBrush*           pBlackBrush = NULL;
ID2D1LinearGradientBrush*       pBackBrush  = NULL;
ID2D1GradientStopCollection*    pGradientStops = NULL;

IWICImagingFactory*             pWICFactory  = NULL;
IDWriteFactory*                 pWFactory   = NULL;
IDWriteTextFormat*              pTextFormat = NULL;
ID2D1Bitmap*                    pIbuttons[4] = {NULL};

IDirectSound*                   pDSnd;
IDirectSoundBuffer*             pDSSndBuffBack;
IDirectSoundBuffer*             pDSSndBuffFront;

WinWave*                        wwBg = NULL;
WinWave*                        wwSelect = NULL;

DSBUFFERDESC                    dsBuffDescBack;
DSBUFFERDESC                    dsBuffDescFront;

HANDLE                          hDThread = INVALID_HANDLE_VALUE;
RECT                            winRect = {0};
POINT                           msdragpt = {0};

////////////////////////////////////////////////////////////////////////////////

static int  msx         = 0;    /// Mouse Coordination X
static int  msy         = 0;    /// Mouse Coordination Y
static int  mslb        = 0;    /// Left Mouse Button.
static int  msrb        = 0;    /// Right Mouse Button.

const int   refWidth    = 1920; /// It will use FHD resolution.
const int   refHeight   = 1080; /// It will use FHD resolution.

HWND g_hwnd = NULL;

////////////////////////////////////////////////////////////////////////////////

LRESULT CALLBACK DlgMsgLoop(HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam);

bool CreateMainDlg(int w,int h,TCHAR *ClassName,void* DlgMsgLoop,void* hInst);
void ErrMsgOut(HWND hwnd,const TCHAR *str);

void DDDestroy();
void DDFlipBack(bool vw);
bool LoadPNG2DDBitmap( const TCHAR* pngfn, ID2D1Bitmap* &out );
bool AllocDSound( IDirectSoundBuffer* &dsbuff, WinWave* ww, DSBUFFERDESC &dsdesc );

void ClearD2Display(unsigned long Color);

DWORD WINAPI drawThread( LPVOID p );

////////////////////////////////////////////////////////////////////////////////

bool DDCreate(HWND hwnd,int w,int h)
{
    HRESULT hr = E_FAIL;

    hr = D2D1CreateFactory( D2D1_FACTORY_TYPE_SINGLE_THREADED, &pD2DFactory );

    if ( hr == S_OK )
    {
        HDC dxDC = GetDC(g_hwnd);

        if ( dxDC == NULL )
        {
            return false;
        }

        SetLayout( dxDC, LAYOUT_BITMAPORIENTATIONPRESERVED );

        D2D1_RENDER_TARGET_PROPERTIES dxProp;

#ifdef USING_DC
        dxProp = RenderTargetProperties( D2D1_RENDER_TARGET_TYPE_DEFAULT,
                                         PixelFormat(DXGI_FORMAT_B8G8R8A8_UNORM,  D2D1_ALPHA_MODE_PREMULTIPLIED ),
                                         0,
                                         0,
                                         D2D1_RENDER_TARGET_USAGE_NONE,
                                         D2D1_FEATURE_LEVEL_DEFAULT );

        hr = pD2DFactory->CreateDCRenderTarget( &dxProp, &pD2DTarget );
#else
        dxProp = RenderTargetProperties();
        dxProp.usage = D2D1_RENDER_TARGET_USAGE_GDI_COMPATIBLE;

        RECT rc;
        GetClientRect( hwnd, &rc );
        D2D1_SIZE_U dsrcsize = SizeU( rc.right- rc.left, rc.bottom - rc.top );

        hr = pD2DFactory->CreateHwndRenderTarget( dxProp,
                                                  HwndRenderTargetProperties( hwnd, dsrcsize ),
                                                  &pD2DTarget );
#endif


        if ( hr == S_OK )
        {
#ifdef USING_DC
            RECT rc;
            GetClientRect( g_hwnd, &rc );
            pD2DTarget->BindDC( dxDC, &rc );
#endif
            pD2DTarget->SetTextAntialiasMode( D2D1_TEXT_ANTIALIAS_MODE_GRAYSCALE );
            pD2DTarget->CreateSolidColorBrush( ColorF(ColorF::White), &pBlackBrush );

            //pBackBrush
            D2D1_GRADIENT_STOP gradientStops[2];
            gradientStops[0].color = ColorF( ColorF::Black, 1);
            gradientStops[0].position = 0.0f;
            gradientStops[1].color = ColorF( ColorF::Blue, 1);
            gradientStops[1].position = 1.0f;
            pD2DTarget->CreateGradientStopCollection( gradientStops,
                                                      2,
                                                      D2D1_GAMMA_2_2,
                                                      D2D1_EXTEND_MODE_CLAMP,
                                                      &pGradientStops );

            pD2DTarget->CreateLinearGradientBrush( LinearGradientBrushProperties( Point2F( refWidth / 2, 0),
                                                                                  Point2F( refWidth / 2, refHeight) ),
                                                   pGradientStops,
                                                   &pBackBrush );

        }

        DWriteCreateFactory( DWRITE_FACTORY_TYPE_SHARED,
                             __uuidof(IDWriteFactory),
                            reinterpret_cast<IUnknown**>(&pWFactory) );

        pWFactory->CreateTextFormat( TEXT("Tahoma"),
                                       NULL,
                                       DWRITE_FONT_WEIGHT_REGULAR,
                                       DWRITE_FONT_STYLE_NORMAL,
                                       DWRITE_FONT_STRETCH_NORMAL,
                                       12.0f,
                                       TEXT("en-us"),
                                       &pTextFormat );

        CoInitialize( NULL );
        CoCreateInstance( CLSID_WICImagingFactory,
                          NULL,
                          CLSCTX_INPROC_SERVER,
                          IID_PPV_ARGS( &pWICFactory ) );

        LoadPNG2DDBitmap( TEXT("testimg01.png"), pIbuttons[0] );
        LoadPNG2DDBitmap( TEXT("testimg02.png"), pIbuttons[1] );
        LoadPNG2DDBitmap( TEXT("testimg03.png"), pIbuttons[2] );
        LoadPNG2DDBitmap( TEXT("testimg04.png"), pIbuttons[3] );


        // ---------------------------------------------------------------------
        // Ok, next : Sound !

        hr = DirectSoundCreate( NULL, &pDSnd, NULL );
        if ( hr == S_OK )
        {
            pDSnd->SetCooperativeLevel( hwnd, DSSCL_NORMAL );

            wwBg = new WinWave();
            if ( wwBg != NULL)
            {
                if ( wwBg->LoadWaveFile( TEXT("background.wav") ) == S_OK )
                {
                    AllocDSound( pDSSndBuffBack, wwBg, dsBuffDescBack );

                    delete wwBg;
                    wwBg = NULL;
                }
            }

            wwSelect = new WinWave();
            if ( wwSelect != NULL )
            {
                if ( wwSelect->LoadWaveFile( TEXT("select.wav") ) == S_OK )
                {
                    AllocDSound( pDSSndBuffFront, wwSelect, dsBuffDescFront );

                    delete wwSelect;
                    wwSelect = NULL;
                }
            }

            if ( pDSSndBuffBack != NULL )
            {
                pDSSndBuffBack->Play( 0, 0, DSBPLAY_LOOPING );
            }

        }

        return true;
    }

    return false;
}

void DDDestroy()
{
    if ( pDSSndBuffBack != NULL )
    {
        pDSSndBuffBack->Stop();
        SafeRelease( &pDSSndBuffBack );
    }

    if ( pDSSndBuffFront != NULL )
    {
        pDSSndBuffFront->Stop();
        SafeRelease( &pDSSndBuffFront );
    }

    SafeRelease( &pDSnd );
    SafeRelease( &pD2DTarget );
    SafeRelease( &pBlackBrush );
    SafeRelease( &pWFactory );
    SafeRelease( &pTextFormat );
    SafeRelease( &pWICFactory );
    SafeRelease( &pD2DFactory );
}

void ClearD2Display(unsigned long Color)
{
    if ( pD2DTarget != NULL )
    {
        //pD2DTarget->Clear( ColorF( ColorF::Black ) );
        D2D1_RECT_F rc = {0, 0, refWidth, refHeight };
        pD2DTarget->FillRectangle( &rc, pBackBrush );
    }
}


void DrawGDIText(int x,int y,TCHAR *str,int r,int g,int b)
{
    if ( pD2DTarget != NULL )
    {
        IDWriteTextLayout*      textLayout  = NULL;
        ID2D1SolidColorBrush*   color       = NULL;

        pD2DTarget->CreateSolidColorBrush( ColorF( r, g, b ), &color );

        pD2DTarget->DrawText( str,
                             _tcslen( str ),
                             pTextFormat,
                             RectF(0,0, refWidth, refHeight),
                             color );

        SafeRelease( &color );
    }
}

bool LoadPNG2DDBitmap( const TCHAR* pngfn, ID2D1Bitmap* &out )
{
    if ( pWICFactory == NULL )
        return false;

    IWICBitmapDecoder*      pImgLoader   = {NULL};
    HRESULT                 hr           = S_OK;

    hr = pWICFactory->CreateDecoderFromFilename( pngfn,
                                                 NULL,
                                                 GENERIC_READ,
                                                 WICDecodeMetadataCacheOnDemand,
                                                 &pImgLoader );
    if ( hr == S_OK )
    {
        IWICBitmapFrameDecode* pWICFdecoder = NULL;
        hr = pImgLoader->GetFrame( 0, &pWICFdecoder );

        if ( hr == S_OK )
        {
            IWICFormatConverter *pFMTconv = NULL;

            hr = pWICFactory->CreateFormatConverter( &pFMTconv );
            if ( hr == S_OK )
            {
                pFMTconv->Initialize( pWICFdecoder,
                                      GUID_WICPixelFormat32bppPBGRA,
                                      WICBitmapDitherTypeNone,
                                      NULL,
                                      0.0f,
                                      WICBitmapPaletteTypeCustom );

                hr = pD2DTarget->CreateBitmapFromWicBitmap( pFMTconv, NULL, &out );
                if ( hr == S_OK )
                {
                    SafeRelease( &pImgLoader );
                    SafeRelease( &pFMTconv );
                    SafeRelease( &pWICFdecoder );
                    return true;
                }
            }
        }

        SafeRelease( &pImgLoader );
    }

    return false;
}

bool AllocDSound( IDirectSoundBuffer* &dsbuff, WinWave* ww, DSBUFFERDESC &dsdesc )
{
    if ( pDSnd == NULL )
        return false;

    if ( ww == NULL )
        return false;

    ZeroMemory( &dsdesc, sizeof( DSBUFFERDESC ) );
    dsdesc.dwSize = sizeof( DSBUFFERDESC );
    dsdesc.dwFlags = DSBCAPS_STATIC | DSBCAPS_CTRLPOSITIONNOTIFY;
    dsdesc.dwBufferBytes = ww->WaveSize();
    dsdesc.lpwfxFormat = ww->WaveFormat();

    HRESULT hr = pDSnd->CreateSoundBuffer( &dsdesc, &dsbuff, NULL );

    if ( hr == S_OK )
    {
        void* pwData1 = NULL;
        void* pwData2 = NULL;
        DWORD lenData1 = 0;
        DWORD lenData2 = 0;

        dsbuff->Lock( 0,
                      dsdesc.dwBufferBytes,
                      &pwData1,
                      &lenData1,
                      &pwData2,
                      &lenData2,
                      0 );

        memcpy( pwData1, ww->WaveData(), dsdesc.dwBufferBytes );
        dsbuff->Unlock( pwData1, dsdesc.dwBufferBytes, NULL, 0 );
    }
}


//////////////////////////////////////////////////////////////

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow)
{
    MSG msg;

    CreateMainDlg( refWidth, refHeight, TEXT("D2D Test"), (void*)DlgMsgLoop, hInstance );

    if(!g_hwnd)
        return FALSE;

    GetWindowRect( g_hwnd, &winRect );

    bool ret = DDCreate( g_hwnd, refWidth , refHeight );
    if(ret==FALSE)
        return FALSE;

    hDThread = CreateThread( NULL, 0, drawThread, NULL, 0, NULL );

    while(1)
    {
        if(PeekMessage(&msg,NULL,0,0,PM_NOREMOVE))
        {
            if(!GetMessage(&msg,NULL,0,0)) return msg.wParam;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
        else
        {
            WaitMessage();
        }
    }

    SuspendThread( hDThread );

    DDDestroy();
    ShowCursor(TRUE);
    return 1;
}

void ErrMsgOut(HWND hwnd, const TCHAR *str)
{
    MessageBox(hwnd,str,TEXT("Error."),MB_OK);
    DestroyWindow(hwnd);
}

bool CreateMainDlg(int w,int h,TCHAR *ClassName,void* DlgMsgLoop,void* hInst)
{
    WNDCLASS wc;

    wc.style            = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc      = (long (__stdcall *)(struct HWND__ *,unsigned int,unsigned int,long))DlgMsgLoop;
    wc.cbClsExtra       = 0;
    wc.cbWndExtra       = 0;
    wc.hInstance        = (HINSTANCE)hInst;
    wc.hIcon            = LoadIcon((HINSTANCE)hInst,(const TCHAR *)IDI_APPLICATION);
    wc.hCursor          = LoadCursor(NULL,IDC_ARROW);
    wc.hbrBackground    = NULL;
    wc.lpszClassName    = ClassName;
    wc.lpszMenuName     = NULL;

    RegisterClass(&wc);

    g_hwnd = CreateWindowEx( 0,
                             ClassName,
                             ClassName,
                             WS_POPUPWINDOW,
                             0,
                             0,
                             refWidth,  //GetSystemMetrics(SM_CXSCREEN),
                             refHeight, //GetSystemMetrics(SM_CYSCREEN),
                             NULL,
                             NULL,
                             (struct HINSTANCE__ *)hInst,
                             NULL );

    if( g_hwnd == NULL )
        return FALSE;

    ShowWindow(g_hwnd,1);
    UpdateWindow(g_hwnd);
    SetFocus(g_hwnd);
    ShowCursor(FALSE);

    return TRUE;
}


/////////////////////////////////////////////////////////////////

bool msdragging = false;

LRESULT CALLBACK DlgMsgLoop(HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam)
{
    switch(message)
    {
        case WM_SIZE:
            {
                UINT width = LOWORD(lParam);
                UINT height = HIWORD(lParam);

                if ( pD2DTarget != NULL )
                {
                    D2D1_SIZE_U size = SizeU( width, height );
#ifndef USING_DC
                    pD2DTarget->Resize( size );
#endif
                }

                if ( msdragging == false )
                {
                    SetWindowPos( hwnd,
                                  NULL,
                                  winRect.left ,
                                  winRect.top,
                                  width,
                                  height,
                                  0 );
                }
                InvalidateRect(hwnd, NULL, FALSE);

                return 0;
            }
            break;

        case WM_DISPLAYCHANGE:
            InvalidateRect(hwnd, NULL, FALSE);
            return 0;
            break;

        case WM_MOUSEMOVE:
            msx=LOWORD(lParam);
            msy=HIWORD(lParam);

            if ( ( mslb == 0 ) && (msdragging == true ) )
            {
                POINT mspt;
                GetCursorPos( &mspt );

                int newx = winRect.left + mspt.x - msdragpt.x;
                int newy = winRect.top + mspt.y - msdragpt.y;

                SetWindowPos( hwnd,
                              NULL,
                              newx,
                              newy,
                              refWidth,
                              refHeight,
                              0 );
            }
            break;

        case WM_LBUTTONDOWN:
            {
                mslb=1;
                if ( msy < 30 )
                {
                    mslb = 0;
                    GetWindowRect( hwnd, &winRect );
                    GetCursorPos( &msdragpt );
                    SetCapture( hwnd );
                    msdragging = true;
                }
            }

            break;

        case WM_RBUTTONDOWN:
            msrb=1;
            if ( msdragging == true )
            {
                msdragging = false;
                ReleaseCapture();
            }
            break;

        case WM_LBUTTONUP:
            mslb=0;
            if ( msdragging == true )
            {
                msdragging = false;
                ReleaseCapture();
            }
            break;

        case WM_RBUTTONUP:
            msrb=0;
            if ( msdragging == true )
            {
                msdragging = false;
                ReleaseCapture();
            }
            break;

        case WM_KEYDOWN:
        case WM_DESTROY:
            {
                DDDestroy();
                ShowCursor(TRUE);
                PostQuitMessage(0);
            }
            break;
    }

    return DefWindowProc(hwnd,message,wParam,lParam);
}

void calcimgputrc( D2D1_RECT_F* rcf, int x, int y, int w, int h )
{
    if ( rcf != NULL )
    {
        rcf->left   = float( x );
        rcf->top    = float( y );
        rcf->right  = rcf->left + float( w );
        rcf->bottom = rcf->top  + float( h );
    }
}

bool checkmouseinrc( D2D1_RECT_F* rcf )
{
    float msxf = float( msx );
    float msyf = float( msy );

    if ( ( msxf > rcf->left ) && ( msxf < rcf->right ) )
        if ( ( msyf > rcf->top ) && ( msyf < rcf->bottom) )
            return true;

    return false;
}

DWORD WINAPI drawThread( LPVOID p )
{
    static int      putwh = 700;
    static float    oppa  = 0.0f;
    static int      putgap = 30;

    static int      dstx  = ( refWidth - ( ( 300 + putgap ) * 4 ) ) / 2;
    static int      dsty  = ( refHeight - 300 ) / 2 ;

    int putx = - ( dstx * 2 );
    int puty = - ( dsty * 2 );

    while( true )
    {
        int startx = putx;
        int starty = puty;
        float oppa_bonus = 0.0f;

        if ( pD2DTarget != NULL )
        {
            pD2DTarget->BeginDraw();

            ClearD2Display( 0 );

            D2D1_RECT_F imgputrc = {0};

            calcimgputrc( &imgputrc, startx, starty, putwh, putwh );
            if ( pIbuttons[0] != NULL )
            {
                if ( checkmouseinrc( &imgputrc ) == true )
                {
                    oppa_bonus += 0.2f;

                    if ( ( mslb == 1 ) && ( pDSSndBuffFront != NULL ) )
                    {
                        pDSSndBuffFront->Play(0,0,0);
                    }
                }

                pD2DTarget->DrawBitmap( pIbuttons[0], imgputrc, oppa + oppa_bonus, D2D1_BITMAP_INTERPOLATION_MODE_LINEAR, NULL );


                if ( oppa_bonus > 0.0f )
                    oppa_bonus = 0.0f;
            }

            startx += putwh + putgap;

            calcimgputrc( &imgputrc, startx, starty, putwh, putwh );
            if ( pIbuttons[1] != NULL )
            {
                if ( checkmouseinrc( &imgputrc ) == true )
                {
                    oppa_bonus += 0.2f;

                    if ( ( mslb == 1 ) && ( pDSSndBuffFront != NULL ) )
                    {
                        pDSSndBuffFront->Play(0,0,0);
                    }
                }

                pD2DTarget->DrawBitmap( pIbuttons[1], imgputrc, oppa + oppa_bonus, D2D1_BITMAP_INTERPOLATION_MODE_LINEAR, NULL );

                if ( oppa_bonus > 0.0f )
                    oppa_bonus = 0.0f;
            }

            startx += putwh + putgap;

            calcimgputrc( &imgputrc, startx, starty, putwh, putwh );
            if ( pIbuttons[2] != NULL )
            {
                if ( checkmouseinrc( &imgputrc ) == true )
                {
                    oppa_bonus += 0.2f;

                    if ( ( mslb == 1 ) && ( pDSSndBuffFront != NULL ) )
                    {
                        pDSSndBuffFront->Play(0,0,0);
                    }
                }

                pD2DTarget->DrawBitmap( pIbuttons[2], imgputrc, oppa + oppa_bonus, D2D1_BITMAP_INTERPOLATION_MODE_LINEAR, NULL );

                if ( oppa_bonus > 0.0f )
                    oppa_bonus = 0.0f;
            }

            startx += putwh + putgap;

            calcimgputrc( &imgputrc, startx, starty, putwh, putwh );
            if ( pIbuttons[3] != NULL )
            {
                if ( checkmouseinrc( &imgputrc ) == true )
                {
                    oppa_bonus += 0.2f;

                    if ( ( mslb == 1 ) && ( pDSSndBuffFront != NULL ) )
                    {
                        pDSSndBuffFront->Play(0,0,0);
                    }
                }

                pD2DTarget->DrawBitmap( pIbuttons[3], imgputrc, oppa + oppa_bonus, D2D1_BITMAP_INTERPOLATION_MODE_LINEAR, NULL );

                if ( oppa_bonus > 0.0f )
                    oppa_bonus = 0.0f;
            }

            if ( putwh > 300 )
            {
                putwh -= ( putwh - 300 ) / 20;
                if ( putwh < 300 )
                {
                    putwh = 300;
                }
            }

            if ( oppa < 0.8f )
            {
                oppa += 0.01f;
            }

            pD2DTarget->SetTransform( Matrix3x2F::Identity() );
            pD2DTarget->DrawEllipse( Ellipse( Point2F( float( msx ), float( msy ) ),
                                                     10.0f,
                                                     10.0f),
                                                     pBlackBrush,
                                                     2.0f );

            TCHAR infostr[1024] = {0};

            _stprintf( infostr,
                       TEXT("Mouse X: %d, Mouse Y: %d"),
                       msx,
                       msy );

            DrawGDIText( 10,10, infostr , 255,255,255 );

            pD2DTarget->EndDraw();
        }

        if ( putx < dstx )
        {
            putx += ( dstx - putx ) / 10;
        }
        else
        if ( putx > dstx )
        {
            putx -= ( dstx - dstx ) / 10;
        }

        if ( puty < dsty )
        {
            puty += ( dsty - puty ) / 10;
        }
        else
        if ( puty > dsty )
        {
            puty -= ( puty - dstx ) / 10;
        }

        Sleep(10);
    }

    return 0;
}
